local LCCC = LibCodesCommonCode

if (LibAchievementsArchive) then return end
local Public = { }
LibAchievementsArchive = Public


--------------------------------------------------------------------------------
-- Initialization
--------------------------------------------------------------------------------

local Internal = {
	name = "LibAchievementsArchive",

	scanThrottle = 500, -- 0.5s

	server = LCCC.GetServerName(),
	userId = GetDisplayName(),
	charId = GetCurrentCharacterId(),

	cache = { },
}
LibAchievementsArchiveInternal = Internal

local API_CUTOFF = 101033
local MAX_ACHIEVEMENT_ID = 3219

local function OnAddOnLoaded( eventCode, addonName )
	if (addonName ~= Internal.name) then return end

	EVENT_MANAGER:UnregisterForEvent(Internal.name, EVENT_ADD_ON_LOADED)

	-- Initialize data store
	if (not LibAchievementsArchiveData and GetAPIVersion() < API_CUTOFF) then
		LibAchievementsArchiveData = { }
	end
	Internal.data = LibAchievementsArchiveData or { }
	if (not Internal.data.characters) then Internal.data.characters = {} end
	Internal.data.keys = nil -- Delete deprecated data

	-- Listen for achievement updates
	if (GetAPIVersion() < API_CUTOFF) then
		LCCC.RunAfterInitialLoadscreen(function( )
			EVENT_MANAGER:RegisterForEvent(Internal.name, EVENT_ACHIEVEMENTS_UPDATED, Internal.Refresh)
			EVENT_MANAGER:RegisterForEvent(Internal.name, EVENT_ACHIEVEMENT_AWARDED, Internal.Refresh)
			EVENT_MANAGER:RegisterForEvent(Internal.name, EVENT_ACHIEVEMENT_UPDATED, Internal.Refresh)
			Internal.Refresh()
		end)
	end
end

EVENT_MANAGER:RegisterForEvent(Internal.name, EVENT_ADD_ON_LOADED, OnAddOnLoaded)


--------------------------------------------------------------------------------
-- Utility Functions
--------------------------------------------------------------------------------

local function NeededBits( n )
	return zo_floor(math.log(n) / math.log(2)) + 1
end

local function Pack( str )
	return LCCC.Chunk(LCCC.Implode(str))
end

local function Unpack( str )
	return LCCC.Explode(LCCC.Unchunk(str))
end


--------------------------------------------------------------------------------
-- Criteria Keys
--------------------------------------------------------------------------------

local CRITERIA_FIELD_SIZE = 5
local CRITERIA_KEYS_RAW = {
	-- This data was generated by the ScanKeys function, which could be found in the code for LibAchievementsArchive version 1.0.0
	"~0A3^1^1^1^1^1^1^1^1~04A1~04o10001a1000Fe~051^01^1^001~04r1^1^03000Fe0007q0001a60007q0007q000380001a0001a000381000Fe70007q0007q0007q000380001a0007q~04o4000Fe000Fe000Fe0007q~085^1~04o10001a10003w10007q1000Fe1~04y1~04g1~04g1~04a5^001~04A1~04o10001a1000Fe1002SG~093^3^4^3^~061~04A1~04A1~04A1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1~04A1^1~04A1^1~0451~04A1^~0B1^103q9010J4j015zU40103q9010J4j015zU4010001a0001^1^1^1^1^1^1^1^1^1^1^1^01^1^1^1^1^1^~0C6^6^6^06^6^001^1^1^1^1^1^1^1^1^6^F^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^~0C1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^6^03^01^1^1^1^1^1^1^1^1^1^1^1^1^6^1^1^1^1^1^1^2^06^1^1^6^6^2^01^1^1^1^1^1^1^1^1^1^1^1^1^1^1^G^00010001a01~04A2^006^1^1^1^1^1^1^1^1^1^001^01^3^01^1^1^1^1^1^1^1^1^1^1^1^3^001^1^1^1^1^1^05^5^4^4^1^1^7^7^6^6^7^7^1^1^1^1^1^1^1^1^1^1^5^5^3^03^05^5^G^3^1^1^1^1^1^1^3^1^3^3^G^3^01^6^1^3^1^3^01^1^3^01^6^1^1^1^1^1^1^6^1^1^1^1^1^1^6^1^1^1^1^1^1^1^1^01^1^01^6^1^1^1^1^1^1^3^1^1^01^3^1^01^1^1^1^C^C^C^C^C^5^C^C^C^C^C^5^C^C^C^C^C^5^C^C^1^1^1^8^1^1^1^1^1^1^~091^1^1^00G^1^1^1^~051~0451^2^~091^1^1^1^1^1^1^1^1^3^0G^1^01^3^01^G^1^G^G^1^1^1^1^1^1^1^1^01^6^04^06^1^1^1^1^1^1^F^~04F^F^A^4^01^1^1^01~04x01~04t1~04p1^1^1^1~04h01~04j1~04z1~04b1^1^1^1~04c01~04e1~04e1^1^1^01~04T5^5^1~04o10001a10003w5^005^5^4^4^~0n3^1^01^G^G^G^G^G^G^G^G^010001aF^F^F^G^1^1^1^1^1^1^5^5^G^1^1^1^1^1000Fe1000Fe1002SG01^1^01~0421~04A10001a10001a10001a10001a10001a10001a10001a10001a10001aC^003^2^2^2^3^2^1^1^2^2^5^2^3^1^3^2^2^1^1^103q9010c9Q0103q9010c9Q0103q9010c9",
	"Q~04A^A^A^A^A^000I^5^5^7^3^2^7^3^2^7^001^E^E^H^~065^~0h7^7^7^7^1^0C^7^7^6^8^7^7^8^2^4^5^7^10003810006f10001u10004i10003S1000111000381000CW10002C1~04y10005U1~04e7^7^7^7^7^L^0F^1^F^J^F^1^03^03^06^01^01^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^9^1^1^1^1^1^1^5^006^000C^~0I1^3^3^010003w1~04o1^1^3^2^3^3^3^3^3^3^3^3^4^3^3^3^3^3^10001a10001a10001a10001a2^2^2^2^2^2^6^2^2^2^2^2^2^2^2^2^2^2^2^6^6^2^C^2^2^0O^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1~04o10001a1000Fe1~04U10004i1000kuG^0E^1~04A1~04o10001a09^1^1^1^1^1^1^0004^G^1^5^4^I^2^04^6^6^7^6^4^6^6^7^6^6^7^6^4^6^6^7^0005^5^5^7^8^5^5^5^001^1^1^1^C^01^2^2^4^4^4^4^4^4^4^4^4^4^4^4^4^4^4^4^4^4^10004i1~04o1^1^C^01^1^1^1^103q90103q90103q90004^1^4^1^1^2^2^6^1^1^1~04o10004i6^1^001^1^1^3^1^01^2^E^6^1^1~04P10002M10003w1^1~04A1~04o10001a1^1~04A1~04o10001a1^8^D^~086^06^1~04P0C^7^1^1~04o10001a1^0E^1^04^1^C^0001^10001a1000Fe1^10001a2^200OQW00OQW1000Fe1^1~04A10001a100OQW103q9010001a1000Fe1002SG1^1~04o10001a1^1^1^1^1^3^1^1^1^1^1^1^1^1^1^1^1~04o10001a1^1^1^1^1^1~04A10001a10007q1^4^4^1^5^5^1~0431~0491~04F1~04L1~04R01~0456^6^K^~062^004^1^1~04o10001a00010001a1^1~04o10001a1~04o10001a1^01^1^10001a10001B1^1^3^0001^1^1^1^1^1^~061^1^1^1^1^1^1^9^9^1^2^0001000Fe1002SG100OQW1^1^1~04F5^E^E^H^6^6^2^K^1^1^1^1~04e01^1^01^1~0471~04U1^1~0471~04U1^C^E^6^1^1^6^4^6^E^7^7^C^1^1^2^1^1^2^4^2^1^1^1^1^1~04C5^1~04G1^1^1^1^1^1^1~0471~04F1^1^1^1^5^5^5^1^4^01~04P1^01^1^1^5^1^1^4^1^1~0451~04F1^1~04A1~04U4^1^1^1~0455000#W000#W000#W000#W000#W5000#W000#W000#W000#W000#W5000#W000#W000#W000#W000#W1000#W4^G^E^E^1^E^E^E",
	"^E^1^1^1^1^E^E^E^2^1^1^2^2^6^C^1~04E1~04H5^1^1^1^1^1^1^1^1^1^1^1^1~0471~04F1~0451^L^1^1~04A1~04U1^1~04K1~04K3^7^1^1~04P10001a1^6^8^4^4^4^4^4^4^4^4^01^~0S1^4^4^1^1^1^1~04o10009O4^1^1^1^1^1^1^1^1^1^05^5^1^1^1^1~04o10009O4^1^1^1^1^1^1^1^02^1^1^1~04o10001a1^0E^4^03^5^10005U10001a1^1^1^6^5^1000721~04o1^1^1^6^1^1^5^10009O10005U1^1^1^6^1^6^10006G10001a1^1^1^6^1^6^10002l10004i1^1^1^6^1^6^1000CW10007q1^1^1^6^1^6^10008c1~04y1^1^1^3^6^10006G10002l1^1^1^6^10006G10006G1^1^1^6^6^10009O10008c1^1^1^6^10005510005U1^1^1^6^10003w10004i1^1^1^6^10003w10003w1^1^1^6^10007q10002M1^1^1^6^10007q10006G1^1^1^8^10002M1000Bk1^1^1^E^E^E^E^5^5^5^4^3^2^3~045~045~0451^1^1^1^9^9^E^C^1^1~04A1~04U1^1~04A1~04U1^1~04A1~04U1^1~04P10001a6^6^1~04o10004i1^1^1^2~045~0455^5^1~04y10002M1^1^1^1000Fe01^1~04A1~04o10001a0U^E^E^E^3^1^1~04K1~04o3^4^06^8^D^K^1^1^1^3^1^1^1^1^1^1^1^1^1^1^1^1^02^1^1^1^1^1^1^1^103q9010c9Q012#l201^1~04o1000381^1~04P10001a1^1~04P10001a1^1~04A1~04P1^1~04o1^1~04P10001a1~04F01^1~04P10001a1^1~04P10001a1^1~04P10001a1^1~04P10001a1^1~04P10001a1^1~04P10001a1^1~04P10001aE^E^E^E^1^1~04P10001a1^1^1^1^1^1^1^1^1^1^D^D^K^K^1^1^1^1^1^1^1^1^a^1^7^7^P^1^10001a1^1^1^1^5~045~045~045~045~0457^1^1^1^1^4^10001aI^6^I^1^1^6^1^1^1^4^5^5^1^5^5^1^1^1^1^1^1^6^9^2^1~04W7^1^1^1^1^1^1^1^1^1^1~04A1~04U7^7^C^1^1^1^1~04A1~04P1~04o1^1^1^7^1~04A10003w1000Fe1^1~04K10001a1^1~04K10001a1^1~04o10003w1~04A10001a10007q1^1~04A1~04o1001E8100CDG100z2G1^1^1^1^1^1^1^1^1^1^1^1^01^1~04A10001a00E^E^E^E^1^2^001^4^4^004^1^1^1^1^1^1^10001a1^1^1^10c9Q01^E^5^5^10007q10005U1^1^1^4^1^1^1^1^5^",
	"1^1^1^5^5^1~04o10001u1^1^1^4^1^1^1^1^1^1^1^1^1^~091^1^1^1^1^1^6^A^02^~051^1^2^2^6^E^E^E^E^1000d41002SGC^1^~071^1^1^1^1^1^6^1^E^E^1^2^1^5^1^~071^1^1^1^1^1^1^1^1~0471~04U1~04K1~04O1^1~0471~04U1^10001a5^1^1^1^1^1^1^1^1^1^1^1^1^1^1~04P10001z1^1^5^5^1^1^E^E^K^6^A^1^2^1^~041^1~04A1~04o10001a1^1^~05E^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^1^A^~091000381~04o5^5^1^1^1^1^4^1^10004i1~04o5^5^1^1^1^4^1^02^5^~061~04A1~04P1~04o5^5^1~0451^1^E^E^E^E^E^C^07^9^1^1~04A1~04U1^1~04A1~04U1^1^1^1^1^1^1^1^1~04X1^1^1^1^1^1^4^1^1~04P10001a1^1^1^9^9^1^7^9^1~04A1~04P1~04o1^1^1~04K1^1~04F1~04K1^1^03^1^1^1^1^1~04F1~04P1^C^1^1^1^1^1^1^1^1^1^1000381~04K5^5^1^1^1^1^4^1^10005U10002M5^5^1^1^1^4^1^2^~07E^1^1^2^2^6^6^2^1^1^C^01^1^1^1^1^3^1^1^1^1~04A1^1^1^1^1^1^1~04A10001a9^8^E^E^E^C^4^2^1^1~04e10002M~045^5^1~04K1^1^1^1^1^1^1^1^5^10001a5^1^1~04A1~04U1^1~04A1~04U1^1~04A1~04U1^1^3^06^6^E^E^E^5^5^1^1^1^1^1^1^1^1^1^1^1^1^1^01^1~04C1^1^1^1~04K1^1~04K3^1^1^1^1^1^1^1^1~0451^1^1^1^1^1^1^6^7^2^1^1^1^1^1^1^6^C^1000381~04P5^5^1^1^1^1^4^1^10006G1~04o5^5^1^1^1^4^1^2^1^1^1^1^10001a10003805^5^5^5^1^1^2^1^1~04A1~04U001~04A3^1~04A3^001~04A00I^D^H^1~0451~0451^1^A^1^1^2^10001a1~04A1^1~04P1~04o~051^1^1^1^1^1^1^1^1^1~04A1~04U1^1~04A1~04U1^1~04A1~04U1^1^1^0E^E^E^E^E^1~04W7^1^1^06^7^7^2^4^01~04A1~04A1^8^1^1^30001a0001a0001a~08C^001000381~04j5^5^1^1^1^1^4^1^1000381~04A5^5^1^1^1^4^1^2^1^1^2^2^6^6^2^1^1^C^2^01^1^1^1^1^1^1~0431^10002M1^1^1^1^1^1^1^1~04D1^1~04S1^2^2^1~04A0003^1^1^1^1^1^01^1^1^01^1~04A1~04U1~0441~04C1~04K1~04U1^1~04A1~04U01~04G1~04PA^1^4^05^01^1^0E^E^E^1^02^1^1^1~",
	"04P1~04o1^1^1^1^1^1^1^6^6^2^1^1^1^1^1^1^6^C^30001a~04P~04P01^02^~08J^1^1~04o1^1^1^1^1^1^1^1^1^1^1^1^1^1^0I^A^1^1~04A1~04U1000381~04o3^3^01^1^1^4^1^1000381~04o5^5^1^1^1^4^1^2^7^1~04U1^1^7^5^5^1^1^1^1^1^1^1^1^1~04A1~04U1^1~04A1~04U1^1^1^1^1^1^01^1^10002l10001a1^1^8^D^E^E^E^E^2^01^1^1^0E^3^5^1^E^E^E^4^6^1^1^1^1~04A1~04U6^1^E^3^1^E^1~04A10001a1^1^3^1^1~04A10001a01^1~0431~04U10001a04^1^1~0431~04U10001a1^0001^1~0451~04o10002M1^4^1^1^1^1~04U1~04o1^01^1000161^1^1^1^1~0451~04o10002M1^1^1^15zU401^1^1^1000381000113^3^1^1^1^1^4^1^1000381~04U5^5^1^1^1^4^1^2^E^E^4^1^1^2^2^6^6^2^1^1^C^2^02^2^~0D1^1^1^1^1^1~0451^1^1^1^10001a1^~0CE^0E^1^7^7^1^1^1^1^B^1^1^1~04K1~04K1^1~04K1^1002SG10009O10007q6^3^E^1^10001a1^1^1^1^1^1^1^1^1^E^1^05^1^1^1^1^1~04A1~04U1^1~04A1~04U1^1~04A1~04U1^1~04U1~04A1~04C4^E^5^0001~0451^1^1^1^1^1^6^6^2^1^1^1^1^1^1^6^C^I^0E^1^1^1^1^10004i10001aE^2^01^1^5^5^E^E^E^E^E^1^C^1^1^1^1^1^1^1^1~04F01~04o1000383^3^1^1^1^1^4^1^1000381~04o3^3^1^1^1^4^1^2^2^1^1^1^1^1^1^1^1~04A1^1000381^6^6^1^1^1^1^1^1^1^7^1~04Q1^1~04A1~04U1^1~04A1~04U3^3^5^5^3^3^4^1^1^1^1^5^2^7^1^1~04A1~04U1^4^1~0445^5^01^1^1^1^1^1^1^6^E^2^1^E^0100z2G100OQW2^~0X1^1^2^2^6^6^2^1^1^C^4^~0i1^1^1~04A1~04U1^1~04A1~04U3^1^1~04A1~04U1^1~04A1~04U1~0434^3^3^1~04E1^1^1^1^1^1^1^1^1^1^1^",
}

function Internal.GetKey( achievementId )
	if (not Internal.cache.keys) then
		Internal.LoadKeys()
	end

	if (type(achievementId) == "number") then
		return Internal.cache.keys[achievementId] or { }
	else
		return { }
	end
end

function Internal.LoadKeys( )
	Internal.cache.keys = { }
	local encoded = Unpack(CRITERIA_KEYS_RAW)
	local pos = 1

	for id = 1, MAX_ACHIEVEMENT_ID do
		local criteria = { }
		local count = LCCC.Decode(zo_strsub(encoded, pos, pos))
		pos = pos + 1

		if (count > 0) then
			if (zo_strsub(encoded, pos, pos) == "^") then
				pos = pos + 1
				for i = 1, count do
					table.insert(criteria, 1)
				end
			else
				for i = 1, count do
					table.insert(criteria, LCCC.Decode(zo_strsub(encoded, pos, pos + CRITERIA_FIELD_SIZE - 1)))
					pos = pos + CRITERIA_FIELD_SIZE
				end
			end
		end

		Internal.cache.keys[id] = criteria
	end
end


--------------------------------------------------------------------------------
-- Scanning and Encoding
--------------------------------------------------------------------------------

local BITS = 6
local TIMESTAMP_BYTES = 6
local SIZE_PROGRESS_MASK = 0x1F
local SIZE_TIMESTAMP_FLAG = 0x20
local MAX_SAFE_PROGRESS_BYTES = 7

function Internal.Refresh( )
	EVENT_MANAGER:UnregisterForUpdate(Internal.name)
	EVENT_MANAGER:RegisterForUpdate(
		Internal.name,
		Internal.scanThrottle,
		function( )
			EVENT_MANAGER:UnregisterForUpdate(Internal.name)
			Internal.ScanAchievements()
		end
	)
end

function Internal.ScanAchievements( )
	if (GetAPIVersion() >= API_CUTOFF) then return end

	local results = { }

	for id = 1, MAX_ACHIEVEMENT_ID do
		local count = GetAchievementNumCriteria(id)
		local progress = ""
		local sum = 0
		local bitsUsed = 0

		local store = function( bytesToStore )
			local bitsToStore = bytesToStore * BITS
			local sumToStore = BitAnd(sum, 2 ^ bitsToStore - 1)
			sum = BitRShift(sum, bitsToStore)
			bitsUsed = bitsUsed - bitsToStore
			return LCCC.Encode(sumToStore, bytesToStore)
		end

		-- Store full multiple-of-6-bit blocks of information
		for i = 1, count do
			local _, cur, max = GetAchievementCriterion(id, i)
			sum = sum + BitLShift(cur, bitsUsed)
			bitsUsed = bitsUsed + NeededBits(max)
			if (bitsUsed >= BITS) then
				progress = store(zo_floor(bitsUsed / BITS)) .. progress
			end
		end

		-- Finalize: store any remaining bits and strip off any leading zeros
		progress = store(zo_ceil(bitsUsed / BITS)) .. progress
		progress = string.gsub(progress, "^0+", "")

		-- lower 5 bits: size of progress field
		-- upper bit: 1 for a 6-byte timestamp, 0 for a zero timestamp
		local sizeFlag = zo_strlen(progress)

		-- For large progress that exceed 42 bits, also store the decimal progress
		local decProgress = ""
		if (sizeFlag > MAX_SAFE_PROGRESS_BYTES) then
			decProgress = Id64ToString(GetAchievementProgress(id))
			decProgress = LCCC.Encode(zo_strlen(decProgress), 1) .. decProgress
		end

		-- Acquire the timestamp
		local timestamp = tonumber(Id64ToString(GetAchievementTimestamp(id)))
		if (timestamp > 0) then
			timestamp = LCCC.Encode(timestamp, TIMESTAMP_BYTES)
			sizeFlag = BitOr(sizeFlag, SIZE_TIMESTAMP_FLAG)
		else
			timestamp = ""
		end

		table.insert(results, string.format("%s%s%s%s", LCCC.Encode(sizeFlag, 1), progress, timestamp, decProgress))
	end

	-- Store the data
	Internal.data.characters[Internal.charId] = string.format("%s,%s,%s", Internal.server, Internal.userId, GetUnitName("player"))
	Internal.data[Internal.charId] = Pack(table.concat(results, ""))

	-- Invalidate the cache
	Internal.cache[Internal.charId] = nil
end


--------------------------------------------------------------------------------
-- Decoding
--------------------------------------------------------------------------------

function Internal.DecodeCharacterData( charId )
	if (type(charId) ~= "string" or not Internal.data or not Internal.data[charId]) then
		-- Invalid character
		return nil
	end

	if (Internal.cache[charId]) then
		-- Data already cached
		return Internal.cache[charId]
	end

	-- Decode and cache the data
	local decoded = { }
	local encoded = Unpack(Internal.data[charId])
	local pos = 1

	for id = 1, MAX_ACHIEVEMENT_ID do
		-- Extract the encoded size flag
		local sizeFlag = LCCC.Decode(zo_strsub(encoded, pos, pos))
		pos = pos + 1

		-- Extract the encoded progress and timestamp
		local encodedProgress = ""
		local encodedTimestamp = ""

		if (BitAnd(sizeFlag, SIZE_PROGRESS_MASK) > 0) then
			local oldPos = pos
			pos = pos + BitAnd(sizeFlag, SIZE_PROGRESS_MASK)
			encodedProgress = zo_strsub(encoded, oldPos, pos - 1)
		end

		if (BitAnd(sizeFlag, SIZE_TIMESTAMP_FLAG) > 0) then
			local oldPos = pos
			pos = pos + TIMESTAMP_BYTES
			encodedTimestamp = zo_strsub(encoded, oldPos, pos - 1)
		end

		-- Extract the full progress field
		local progress
		if (zo_strlen(encodedProgress) > MAX_SAFE_PROGRESS_BYTES) then
			local decSize = LCCC.Decode(zo_strsub(encoded, pos, pos))
			pos = pos + 1
			local oldPos = pos
			pos = pos + decSize
			progress = zo_strsub(encoded, oldPos, pos - 1)
		else
			progress = LCCC.Decode(encodedProgress)
		end

		-- Initialize data
		decoded[id] = {
			criteria = { },
			timestamp = LCCC.Decode(encodedTimestamp),
			progress = StringToId64(progress),
		}

		-- Decode the progress into individual criterion numbers
		local bitsUsed = 0
		for _, max in ipairs(Internal.GetKey(id)) do
			local bitPosStart = bitsUsed
			local bitsNeeded = NeededBits(max)
			bitsUsed = bitPosStart + bitsNeeded

			-- Note: bitPos is 0-based, but bytePos is 1-based
			local bytePosStart = zo_floor(bitPosStart / BITS) + 1
			local bytePosEnd = zo_ceil(bitsUsed / BITS)

			local cur = LCCC.Decode(zo_strsub(encodedProgress, -bytePosEnd, -bytePosStart))
			cur = BitRShift(cur, bitPosStart - (bytePosStart - 1) * BITS)
			cur = BitAnd(cur, 2 ^ bitsNeeded - 1)

			table.insert(decoded[id].criteria, cur)
		end
	end

	Internal.cache[charId] = decoded
	return decoded
end

function Internal.ReadData( charId, achievementId )
	local result = {
		criteria = { },
		timestamp = 0,
		progress = StringToId64("0"),
	}

	local key = Internal.GetKey(achievementId)

	if (#key == 0) then
		-- Invalid achievement; do nothing and return the default empty data
	else
		local data = Internal.DecodeCharacterData(charId)
		if (data) then
			-- Valid achievement and character
			result = data[achievementId]
		else
			-- Valid achievement, but invalid character; populate the criteria with zeros
			for i = 1, #key do
				table.insert(result.criteria, 0)
			end
		end
	end

	return result
end


--------------------------------------------------------------------------------
-- Miscellaneous
--------------------------------------------------------------------------------

function Internal.GetMaxAchievementId( )
	return MAX_ACHIEVEMENT_ID
end
